from dataclasses import replace
from logging import getLogger
from time import sleep
import tarfile
import typing
import sys

from zstandard import ZstdCompressor

from ...shared import (
    MAX_SOFT_SIZE_MIN,
    LFS,
    UploadRequestParams,
    UploadResponseHeaders,
    UploadEC,
    SpooledTempFile,
    version,
)
from .errors import MultipleClients, ChannelLocked, ReportThis, VersionError
from .util import wait_delay_sec, request
from .progress import Progress
from .crypt import encrypt
from .io import IO

if typing.TYPE_CHECKING:
    from requests import Response
    from collections.abc import Callable
    from .data import Result, Config, Mode


_LOG = "send"
_DEFAULT_LVL: int = 3


def _send_known_error(r: Response) -> None:
    """
    Raise an exception according to the send response error
    If error is unknown, does nothing
    """
    match r.status_code:
        case UploadEC.illegal_version:
            raise VersionError(r.text)
        case UploadEC.conflict:
            raise MultipleClients("The stream ID changed mid-upload; maybe the receiver broke the pipe?")
        case UploadEC.wrong_version | UploadEC.too_big | UploadEC.forbidden | UploadEC.stream_id:
            raise ReportThis(r.text)
        case UploadEC.locked:
            raise ChannelLocked(r.text)


def _send_block(data: bytes, conf: Config, params: UploadRequestParams, *, lvl: int = 0) -> Response:
    """
    Upload the given block of data; updates params for next block
    """
    typ = "POST" if params.stream_id is None else "PUT"
    r = request(typ, conf.channel_url(), params=params.to_dict(), data=data, timeout=conf.timeout)
    if r.ok:
        return r
    if r.status_code == UploadEC.wait:
        delay = wait_delay_sec(lvl)
        getLogger(_LOG).info("Pipe full, sleeping for %s second(s).", delay)
        sleep(delay)
        return _send_block(data, conf, params, lvl=lvl + 1)
    _send_known_error(r)
    raise RuntimeError(f"Error {r.status_code}", r.text)


def _send_data(
    conf: Config, progress: Progress, io: IO, compress: Callable[[bytes], bytes], params: UploadRequestParams
) -> None:
    """
    Send data to the remote pipe, using the preconfigured parameters provided
    """
    log = getLogger(_LOG)
    while not params.final:
        block, params.final = io.read()
        log.info("Processing block of %s", LFS(block))
        enc = encrypt(block, compress, conf.password)
        r = _send_block(enc, conf, params)
        progress.update(block)
        if params.stream_id is None:  # configure following PUTs
            if params.final:
                return
            headers = UploadResponseHeaders.from_dict(r.headers)
            params.stream_id = headers.stream_id
            io.increase_chunk(headers.max_size)
            sleep(0.025)  # Avoid being over-eager with sending data; let the read thread read


def _send(conf: Config, mode: Mode, file: int | SpooledTempFile) -> Result:
    """
    Send data to the remote pipe reading from file
    """
    log = getLogger(_LOG)
    lvl = _DEFAULT_LVL if mode.zstd is None else mode.zstd
    log.debug("Using compression level %d and %d threads", lvl, mode.threads)
    compress = ZstdCompressor(write_checksum=True, level=lvl, threads=mode.threads).compress
    io = IO(file, MAX_SOFT_SIZE_MIN)
    sleep(0.025)  # Avoid being over-eager with sending data; let the read thread read
    params = UploadRequestParams(
        version=version,
        final=False,
        ttl=mode.ttl,
        encrypted=conf.password is not None,
    )
    log.info("Writing to channel %s", conf.channel)
    with Progress(conf, mode) as progress:
        _send_data(conf, progress, io, compress, params)
    log.info("Stream complete")
    return progress.result


def send(conf: Config, mode: Mode) -> Result:
    """
    Send data to the remote pipe
    """
    log = getLogger(_LOG)
    # Tarball dir
    if mode.dir is not None:
        if not mode.dir.is_dir():
            raise FileNotFoundError(f"Upload directory missing: {mode.dir}")
        # Note: in_f is never a file descriptor to keep the handler alive / file open
        in_f: SpooledTempFile | typing.IO[bytes] = SpooledTempFile()
        log.info("Creating tarball from: %s", mode.dir)
        with tarfile.open(fileobj=in_f, mode="w:gz") as tb:
            tb.add(mode.dir, recursive=True)
        in_f.seek(0)
    else:
        in_f = sys.stdin.buffer if mode.file is None else mode.file.open("rb")
    # Update progress
    if mode.file and not mode.progress:
        size = mode.file.stat().st_size
        log.debug("Setting: --progress %d", size)
        mode = replace(mode, progress=size)
    # Send file
    try:
        return _send(conf, mode, in_f if isinstance(in_f, SpooledTempFile) else in_f.fileno())
    finally:
        if in_f != sys.stdin.buffer:
            in_f.close()



from dataclasses import fields
from pathlib import Path
import argparse

from ..shared import __version__
from .app import ServerConfig, LogConfig, serve
from .util import MIN_VERSION


def cli() -> None:
    parser = argparse.ArgumentParser()
    modes = parser.add_argument_group("Modes")
    modes.add_argument("-V", "--version", action="version", version=f"{parser.prog} {__version__}")
    modes.add_argument(
        "--min-client-version",
        action="version",
        version=f"rpipe>={MIN_VERSION}",
        help="Print the minimum supported client version then exit",
    )
    parser.add_argument("port", type=int, help="The port waitress will listen on")
    parser.add_argument("--host", default="0.0.0.0", help="The host waitress will bind to for listening")
    parser.add_argument("-b", "--blocklist", type=Path, help="The blocklist configuration file")
    parser.add_argument("-s", "--state-file", type=Path, help="The save state file, if desired")
    parser.add_argument(
        "-k",
        "--key-files",
        type=Path,
        nargs="*",
        default=[],
        help="SSH ed25519 public keys to accept for admin access",
    )
    parser.add_argument("-F", "--favicon", type=Path, help="The favicon file, if desired")
    log_g = parser.add_argument_group("Logging")
    log_g.add_argument(
        "-l", "--log-file", type=Path, default=None, help="The log file to append to, if desired"
    )
    # pylint: disable=duplicate-code
    log_g.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase Log verbosity, pass more than once to increase verbosity",
    )
    log_g.add_argument("-C", "--colored", action="store_true", help="Enable color in the log output")
    parser.add_argument("--debug", action="store_true", help="Run the server in debug mode")
    ns = parser.parse_args()
    gen = lambda C: C(**{i: getattr(ns, i) for i in (k.name for k in fields(C))})
    serve(gen(ServerConfig), gen(LogConfig), ns.favicon)



from dataclasses import dataclass, asdict, field
from typing import TYPE_CHECKING, cast
from logging import getLogger
from datetime import datetime
from threading import RLock
import atexit
import json
import re

from werkzeug.serving import is_running_from_reloader
from flask import request

from ..shared import TRACE, Version, version, __version__

if TYPE_CHECKING:
    from pathlib import Path


@dataclass(kw_only=True, slots=True)
class Data:
    """
    Contains blocklist data such as ips, routes, whitelists, etc
    """

    version: Version = field(default_factory=lambda: Version("0.0.1"))
    ip_whitelist: set[str] = field(default_factory=set)
    ip_blacklist: set[str] = field(default_factory=set)
    route_whitelist: list[str] = field(default_factory=list)
    route_blacklist: list[str] = field(default_factory=list)
    stats: dict[str, list[list[str]]] = field(default_factory=dict)


class Blocked:  # Move into server? Move stats into Stats?
    """
    Used to determine if requests should be blocked or not
    """

    __slots__ = ("_log", "_data", "_file", "_lock", "_white_pat", "_black_pat")
    MIN_VERSION = Version("9.11.0")

    def __init__(self, file: Path | None, debug: bool) -> None:
        self._log = getLogger("Blocked")
        if file is not None:
            self._log.info("Loading blocklist: %s", file)
        js: dict
        js = {"version": __version__} if file is None or not file.is_file() else json.loads(file.read_text())
        if (old := Version(js.pop("version", ""))) < self.MIN_VERSION:
            raise ValueError(f"Blocklist version too old: {old} <= {self.MIN_VERSION}")
        js.update({i: set(k) for i, k in js.items() if i.startswith("ip_")})
        self._data = Data(version=version, **js)  # Use new version
        self._white_pat: list[re.Pattern[str]] = []
        self._black_pat: list[re.Pattern[str]] = []
        self._file: Path | None = file
        self._lock = RLock()
        with self as _:
            pass  # Generates patterns
        # Initialize file as needed
        if file is None:
            self._log.warning("No blocklist is set, blocklist changes will not persist across restarts")
            return
        if not file.exists():
            self._log.warning("Blocklist %s not found. Using defaults", file)
        # Setup saving on exit
        if debug and not is_running_from_reloader():  # Flask will reload the program, skip atexit
            self._log.info("Skipping initialization until reload")
            return
        self._log.info("Installing atexit shutdown handler for saving blocklist")
        atexit.register(self._save)

    def _mk_pat_from(self, patterns: list[str]) -> list[re.Pattern[str]]:
        ret = []
        for i in patterns:
            try:
                ret.append(re.compile(i, re.IGNORECASE))
            except re.PatternError:
                self._log.error("Could not compile pattern: %s", i)
        return ret

    def __enter__(self) -> Data:
        """
        Returns the Data object of Blocked
        """
        self._lock.acquire()
        return self._data

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._white_pat = self._mk_pat_from(self._data.route_whitelist)
        self._black_pat = self._mk_pat_from(self._data.route_blacklist)
        self._lock.release()

    def _save(self) -> None:
        """
        Save data to blocklist
        This function assumes self._file is not None
        """
        if self._file is None:
            self._log.critical("_save called when blocklist file is not set; changes will not persist")
            return
        try:
            self._log.info("Saving blocklist: %s", self._file)
            with self as data:
                sstr = lambda x: list(x) if isinstance(x, set) else str(x)
                self._file.write_text(json.dumps(asdict(data), default=sstr, indent=4))
        except OSError:
            self._log.exception("Failed to save blocklist %s", self._file)

    def _notate(self, ip: str) -> None:
        """
        Record the blocked route (should be called by __call__)
        """
        pth = request.path
        self._log.log(TRACE, "%s blocked, attempted path: %s", ip, pth)
        with self as data:
            if ip not in data.stats:
                data.stats[ip] = []
            data.stats[ip].append([str(datetime.now()), pth])

    @staticmethod
    def _match(pth: str, patterns: list[re.Pattern[str]]) -> bool:
        """
        :param pth: The path string to check for matching
        :param patterns: The patterns to test s against
        :return: True iff s matches any of the given patterns
        """
        return any(re.fullmatch(i, pth) is not None for i in patterns)

    def __call__(self) -> bool:
        """
        :return: True if the given request should be blocked
        """
        ip = cast(str, request.headers.get("X-Forwarded-For", request.remote_addr))
        with self as data:
            # Block / allow based on ip first, then routes
            if ip in data.ip_whitelist:
                return False
            if ip in data.ip_blacklist:
                self._notate(ip)
                return True
            pth = request.path
            if self._match(pth, self._white_pat):
                return False
            if self._match(pth, self._black_pat):
                self._log.info("Blocking IP %s based on route: %s", ip, pth)
                data.ip_blacklist.add(ip)
                self._notate(ip)
                return True
        return False



from typing import TYPE_CHECKING
from dataclasses import asdict
from logging import getLogger

from flask import request

from ...shared import TRACE, DeleteEC, QueryEC
from ..util import plaintext, json_response
from ..server import ServerShutdown
from .write import write
from .read import read

if TYPE_CHECKING:
    from flask import Response
    from ..server import State


def _delete(state: State, channel: str) -> Response:
    log = getLogger("delete")
    with state as u:
        u.stats.delete(channel)
        if (s := u.streams.get(channel, None)) is None:
            return plaintext("Channel already gone", status=204)
        if s.locked:
            return plaintext("Channel is locked", status=DeleteEC.locked)
        log.info("Deleting channel %s", channel)
        del u.streams[channel]
    return plaintext("Deleted", status=202)


def _handler(state: State, channel: str) -> Response:
    log = getLogger("channel")
    try:
        match request.method:
            case "DELETE":
                return _delete(state, channel)
            case "GET":
                return read(state, channel)
            case "POST" | "PUT":
                return write(state, channel)
            case _:
                log.warning("404: bad method: %s", request.method)
                return plaintext(f"Unknown method: {request.method}", status=404)
    except ServerShutdown:
        log.warning("Ignoring request, server is shutting down")
        return plaintext("Server is shutting down", status=503)


def handler(state: State, channel: str) -> Response:
    log = getLogger("channel")
    log.debug("Invoking: %s %s", request.method, channel)
    ret = _handler(state, channel)
    if ret.status_code >= 300:
        log.debug("Sending: %s", ret)
        log.log(TRACE, "Body: %s", ret.get_data())
    return ret


def query(state: State, channel: str) -> Response:
    log = getLogger("query")
    log.debug("Query %s", channel)
    with state as u:
        if (s := u.streams.get(channel, None)) is None:
            log.debug("Channel not found: %s", channel)
            return plaintext("No data on this channel", status=QueryEC.no_data)
        q = s.query()
    log.debug("Channel found: %s", q)
    return json_response(asdict(q))



from tempfile import SpooledTemporaryFile
from contextlib import contextmanager
from typing import TYPE_CHECKING
from logging import getLogger
from os import umask

from psutil import virtual_memory
from flask import request

if TYPE_CHECKING:
    from collections.abc import Sequence


def remote_addr() -> str:
    addr = request.remote_addr
    xf = request.headers.get("X-Forwarded-For")
    return f"{addr} / X-Forwarded-For: {xf}" if xf else str(addr)


def total_len(x: Sequence[bytes]) -> int:
    return sum(len(i) for i in x)


@contextmanager
def restrict_umask(mask: int):
    old = umask(0o66)  # Get the old umask
    try:
        yield umask(old | mask)
    finally:
        umask(old)


class SpooledTempFile(SpooledTemporaryFile):
    """
    A subclass of SpooledTemporaryFile that provides a
    method of checking if the files is fully in memory
    """

    _log = "TempFile"

    def __init__(self) -> None:
        n = virtual_memory().available
        getLogger(self._log).debug("Initializing in memory with max_size=%d", n)
        super().__init__(n)
        self._is_virtual: bool = True  # self._rolled is private so we use this instead

    @property
    def is_virtual(self) -> bool:
        """
        :return: True if the file is spooled / contained entirely in memory
        """
        return self._is_virtual

    def rollover(self) -> None:
        if self._is_virtual:
            log = getLogger(self._log)
            log.info("max_size exceeded; rolling onto disk")
            super().rollover()
            log.debug("Rollover complete")
        self._is_virtual = False



from .version_ import __version__, version, Version, WEB_VERSION
from .request_response import (
    MAX_SOFT_SIZE_MIN,
    UploadRequestParams,
    UploadResponseHeaders,
    DownloadRequestParams,
    DownloadResponseHeaders,
    QueryResponse,
    AdminMessage,
)
from .error_code import BLOCKED_EC, UploadEC, DownloadEC, DeleteEC, QueryEC, AdminEC
from .util import restrict_umask, remote_addr, SpooledTempFile, total_len
from .stats import AdminStats, Stats
from .log import TRACE, LFS



from typing import TYPE_CHECKING
from logging import getLogger
import signal
import atexit

from werkzeug.serving import is_running_from_reloader
from zstdlib import Singleton


from .prune_thread import PruneThread
from .state import State

if TYPE_CHECKING:
    from pathlib import Path


_LOG = "Server"


def _ctrlc(sig, *_):
    getLogger(_LOG).error("Received signal %s; raising KeyboardInterrupt", sig)
    raise KeyboardInterrupt()


class Server(Singleton):
    """
    The main server class
    This class is a singleton and should only be instantiated once
    This class is responsible for initializing the server and handling shutdown
    This class may install signal handlers for signal.SIGNALS so that atexit catches them
    """

    SIGNALS = (signal.SIGINT, signal.SIGTERM, signal.SIGQUIT)
    __slots__ = ("state", "_state_file", "_log")

    def shutdown(self):
        """
        Shut the server down
        """
        log = getLogger(_LOG)
        log.critical("Server shutdown initiated")
        with self.state as u:
            if u.shutdown:
                raise RuntimeError("Server already shut down")
            u.shutdown = True
            log.info("Removing atexit shutdown registration")
            atexit.unregister(self.shutdown)
            if self._state_file is not None:
                u.save(self._state_file)
                return
        log.warning("State file not set; state will not be saved")

    def __init__(self, debug: bool, state_file: Path | None) -> None:
        self._log = getLogger(_LOG)
        self._log.info("Initializing server")
        self._state_file: Path | None = state_file
        self.state = State(debug)
        # Flask reloader will just relaunch this so we skip most configuration (such as persistent items)
        if debug and not is_running_from_reloader():
            # It is important not to catch signals in the reloader process
            # It will fail to detect the death of its children and hang
            self._log.info("Skipping initialization until reload")
            return
        if not debug:  # Don't do this for debug, it will be overridden
            self._log.info("Installing signal handlers to ensure graceful shutdown")
            for i in self.SIGNALS:
                signal.signal(i, _ctrlc)
        # Load state file as needed
        if self._state_file is not None:
            with self.state as u:
                u.load(self._state_file)
        # Start prune thread
        self._log.info("Starting prune thread")
        PruneThread(self.state).start()
        # Ensure state is saved on exit
        self._log.info("Installing atexit shutdown handler for state saving")
        atexit.register(self.shutdown)
        self._log.info("Server initialization complete")



"""
This file converts the output of the CLI to the input of the client
It also handles CLI error checking
This is a distinct file from the CLI because importing the client is slow
Splitting the CLI into two files does cause a cyclic import
"""

from logging import StreamHandler, getLevelName, getLogger
from inspect import Parameter, signature
from typing import TYPE_CHECKING
from dataclasses import replace
from os import getenv
import argparse
import sys

from zstdlib.log import CuteFormatter
from human_readable import listing

from ..shared import log
from .client import UsageError, Config, Mode, rpipe
from .cli import PASSWORD_ENV
from .admin import Admin

if TYPE_CHECKING:
    from argparse import Namespace


_LOG = "main"


def _check_mode_flags(mode: Mode) -> None:
    if (mode.recv, mode.send, mode.delete).count(True) != 1:
        raise UsageError("Can only recv, send, or delete at a time")
    # Flag specific checks
    if mode.ttl is not None and mode.ttl <= 0:
        raise UsageError("--ttl must be positive")
    if mode.progress is not False and mode.progress <= 0:
        raise UsageError("--progress argument must be positive if passed")
    if mode.yes and mode.file is None and mode.dir is None:
        raise UsageError("--yes requires --file or --dir")
    # Mode flags
    recv_bad = {"ttl"}
    send_bad = {"block", "peek", "force", "yes"}
    delete_bad = recv_bad | send_bad | {"progress", "encrypt", "file", "dir"}
    bad = lambda x: [f"--{i}" for i in x if bool(getattr(mode, i))]
    fmt = lambda x: f"argument{'' if len(x) == 1 else 's'} {listing(x, ',', 'and') }: may not be used "
    if mode.priority() and (args := bad(delete_bad)):
        raise UsageError(fmt(args) + "with priority modes")
    # Mode specific flags
    if mode.recv and (args := bad(recv_bad)):
        raise UsageError(fmt(args) + "when receiving data from the pipe")
    if mode.send and (args := bad(send_bad)):
        raise UsageError(fmt(args) + "when sending data to the pipe")
    if mode.delete and (args := bad(delete_bad)):
        raise UsageError(fmt(args) + "when deleting data from the pipe")


def _main(ns: Namespace, conf: Config):
    mode_d = {i: k for i, k in vars(ns).items() if i in Mode.keys()}
    # Load mode
    if (ns.file or ns.dir) and not ns.recv and not ns.send:
        raise UsageError("--file and --dir require either an explicit -r or -w")
    if (ns.recv, ns.send, ns.delete).count(True) == 0:
        mode_d["recv"] = ns.file is None and sys.stdin.isatty() and not ns.delete
        mode_d["send"] = not (mode_d["recv"] or ns.delete)
    mode = Mode(**mode_d)
    # Adjustments, error checks, then execute
    _check_mode_flags(mode)
    if ns.encrypt is None:
        mode = replace(mode, encrypt=bool(conf.password))
    if mode.encrypt and not conf.password:
        raise UsageError(f"--encrypt flag requires a password; set via {PASSWORD_ENV}")
    rpipe(conf, mode, ns.config_file)


def _admin(ns: Namespace, conf: Config) -> None:
    kw = signature(func := Admin(conf)[ns.method.replace("-", "_")]).parameters
    assert all(i.kind == Parameter.POSITIONAL_OR_KEYWORD for i in kw.values())
    func(**{i: getattr(ns, i) for i in kw})


def _config_log(parsed: Namespace) -> None:
    # Log config
    log.define_trace()
    root = getLogger()
    root.setLevel(lvl := log.level(parsed.verbose))
    assert len(root.handlers) == 0, "Root logger should not have any handlers"
    root.addHandler(sh := StreamHandler())
    sh.setFormatter(CuteFormatter(**log.CF_KWARGS, colored=not parsed.no_color_log))  # type: ignore[arg-type]
    getLogger(_LOG).info(
        "Logging level set to %s with colors %sABLED",
        getLevelName(lvl),
        "DIS" if parsed.no_color_log else "EN",
    )


# pylint: disable=too-many-locals,too-many-statements
def main(parser: argparse.ArgumentParser, parsed: Namespace) -> None:
    _config_log(parsed)
    # Load Config
    conf_d = {i: k for i, k in vars(parsed).items() if i in Config.keys()}
    if (pw := getenv(PASSWORD_ENV)) is not None:
        getLogger(_LOG).debug("Taking password from: %s", PASSWORD_ENV)
        conf_d["password"] = pw
    conf = Config.load(conf_d, parsed.config_file)  # We do not validate conf yet
    # Invoke the correct function
    if (parsed.method is not None) != parsed.admin:
        raise UsageError("Admin command must be passed with --admin")
    try:
        (_admin if parsed.admin else _main)(parsed, conf)
    except UsageError as e:
        parser.error(str(e))



from logging import DEBUG, INFO, StreamHandler, FileHandler, getLevelName, getLogger, shutdown
from os import environ, close as fd_close
from dataclasses import dataclass
from tempfile import mkstemp
from functools import wraps
from pathlib import Path
import atexit

from flask import Response, Flask, send_file, request
from zstdlib.log import CuteFormatter
import waitress

from ..shared import BLOCKED_EC, TRACE, restrict_umask, remote_addr, log, __version__
from .util import MAX_SIZE_HARD, MIN_VERSION, json_response, plaintext
from .channel import handler, query
from .blocked import Blocked
from .server import Server
from .admin import Admin


_LOG = "app"


@dataclass(frozen=True, slots=True, kw_only=True)
class LogConfig:
    colored: bool
    log_file: Path
    verbose: int
    debug: bool


@dataclass(frozen=True, slots=True, kw_only=True)
class ServerConfig:
    host: str
    port: int
    debug: bool
    state_file: Path | None
    blocklist: Path | None
    key_files: list[Path]


class App(Flask):

    @dataclass(frozen=True, slots=True)
    class Objs:
        admin: Admin
        server: Server
        blocked: Blocked
        favicon: Path | None

    def __init__(self) -> None:
        super().__init__(f"rpipe_server {__version__}")
        getLogger(_LOG).info("Setting max packet size: %s", log.LFS(MAX_SIZE_HARD))
        self.config["MAX_CONTENT_LENGTH"] = MAX_SIZE_HARD
        self.url_map.strict_slashes = False

    def start(self, conf: ServerConfig, log_file: Path, favicon: Path | None):
        lg = getLogger(_LOG)
        if favicon is not None and not favicon.is_file():
            lg.error("Favicon file not found: %s", favicon)
            favicon = None
        blocked = Blocked(conf.blocklist, conf.debug)
        admin = Admin(log_file, conf.key_files, blocked)
        lg.info("Starting server version: %s", __version__)
        # pylint: disable=attribute-defined-outside-init
        self._objs = self.Objs(admin, Server(conf.debug, conf.state_file), blocked, favicon)
        lg.info("Binding to %s:%s", conf.host, conf.port)
        if conf.debug:
            self.run(host=conf.host, port=conf.port, debug=True)
        else:
            waitress.serve(self, host=conf.host, port=conf.port, clear_untrusted_proxy_headers=False)

    def give(self, *, objs: bool = False, logged: bool = True):
        """
        Give the wrapped function self.objs and log requests as requested
        Also handles blocked IP addresses and routes
        """
        lg = getLogger(_LOG)

        def decorator(func):
            @wraps(func)
            def inner(*args, **kwargs):
                if self._objs.blocked():
                    if lg.isEnabledFor(DEBUG):
                        lg.debug("Blocked request by %s to: %s", remote_addr(), request.path)
                    return Response(status=BLOCKED_EC)
                ret = func(*args, self._objs, **kwargs) if objs else func(*args, **kwargs)
                if not logged or self._objs.server.state.debug:
                    return ret
                # Release mode: Log the request before returning it, Flask in debug mode does automatically
                quiet = ret.status_code == 404 and request.full_path.strip("?") == "/favicon.ico"
                lvl = DEBUG if (quiet or ret.status_code in (410, 425) or ret.status_code < 300) else INFO
                lvl = TRACE if request.method in ("OPTIONS", "HEAD") else lvl
                args = (remote_addr(), request.method, request.full_path.strip("?"), ret.status_code)
                lg.log(lvl, '%s - "%s %s" %d', *args)
                return ret

            return inner

        return decorator

    def route(self, *paths, admin: bool = False, objs: bool = False, logged: bool = True, **kwargs):
        """
        Route decorator that allows for multiple paths to be routed to the same function
        Automatically applies the give decorator with objs=objs and logged=logged
        If admin, set objs=True and set methods=["POST"] if not set
        """
        if not paths:
            raise ValueError("At least one path is required")
        if admin and "methods" not in kwargs:
            kwargs["methods"] = ["POST"]
        super_route = super().route

        def wrapper(func):
            ret = self.give(objs=admin or objs, logged=logged)(func)
            for p in paths:
                ret = super_route(p, **kwargs)(ret)
            return ret

        return wrapper


app = App()


#
# Routes
#


@app.errorhandler(404)
@app.give()
def _page_not_found(_, *, quiet=False) -> Response:
    lg = getLogger(_LOG)
    if not quiet:
        lg.warning("404: Not found: %s", request.path)
    (lg.debug if quiet else lg.info)("Headers: %s", request.headers)
    return Response("404: Not found", status=404)


@app.route("/", "/help")
def _help() -> Response:
    msg = (
        "Welcome to the web UI of rpipe. "
        "To interact with a given channel, use the path /c/<channel>. "
        "To read a message from a given channel, use a GET request. "
        "To write a message to a given channel, use PUT and POST requests. "
        "To delete a channel, use a DELETE request. "
        "Note: Using the web version bypasses version consistent checks "
        "and may result in safe but unexpected behavior (such as failing "
        "an uploaded message; if possible use the rpipe client CLI instead. "
        "Install the CLI via: pip install rpipe"
    )
    return plaintext(msg)


@app.route("/favicon.ico", objs=True, logged=False)
def _favicon(o: App.Objs) -> Response:
    return _page_not_found(404, quiet=True) if o.favicon is None else send_file(o.favicon)


@app.route("/robots.txt")
def _robots_txt() -> Response:
    return plaintext("User-agent: *\nDisallow: /\nAllow: /help")


@app.route("/version")
def _show_version() -> Response:
    return plaintext(__version__)


@app.route("/supported")
def _supported() -> Response:
    return json_response({"min": str(MIN_VERSION), "banned": []})


@app.route("/c/<channel>", objs=True, methods=["DELETE", "GET", "POST", "PUT"])
def _channel(o: App.Objs, channel: str) -> Response:
    return handler(o.server.state, channel)


@app.route("/q/<channel>", objs=True)
def _query(o: App.Objs, channel: str) -> Response:
    return query(o.server.state, channel)


# Admin routes


@app.route("/admin/uid", admin=True, methods=["GET"])
def _admin_uid(o: App.Objs) -> Response:
    """
    Get a few UIDSs needed to sign admin requests
    The exact number is up to the server, if you need more, request more
    These UIDs will expire after a short period of time
    """
    return o.admin.uids()


@app.route("/admin/debug", admin=True)
def _admin_debug(o: App.Objs) -> Response:
    return o.admin.debug(o.server.state)


@app.route("/admin/channels", admin=True)
def _admin_channels(o: App.Objs) -> Response:
    return o.admin.channels(o.server.state)


@app.route("/admin/stats", admin=True)
def _admin_stats(o: App.Objs) -> Response:
    with o.blocked as data:
        return o.admin.stats(o.server.state, data.stats)


@app.route("/admin/log", admin=True)
def _admin_log(o: App.Objs) -> Response:
    return o.admin.log(o.server.state)


@app.route("/admin/log-level", admin=True)
def _admin_log_level(o: App.Objs) -> Response:
    return o.admin.log_level(o.server.state)


@app.route("/admin/lock", admin=True)
def _admin_lock(o: App.Objs) -> Response:
    return o.admin.lock(o.server.state)


@app.route("/admin/ip", admin=True)
def _admin_ip(o: App.Objs) -> Response:
    return o.admin.ip(o.server.state)


@app.route("/admin/route", admin=True)
def _admin_route(o: App.Objs) -> Response:
    return o.admin.route(o.server.state)


# Main functions


def _log_shutdown(log_file: Path) -> None:
    getLogger().critical("Logger is shutting down. Purging: %s", log_file)
    shutdown()
    # Missing is an error, but we ignore it since it's not critical and we are shutting down
    log_file.unlink(missing_ok=True)


def _log_config(conf: LogConfig) -> Path:
    rdlf_env: str = "_REUSE_DEBUG_LOG_FILE"
    log.define_trace()
    log_file = conf.log_file
    # Flask debug mode may restart the server without cleaning up, we reuse the log file here
    if reuse_log_file := log_file is None and conf.debug and rdlf_env in environ:
        log_file = Path(environ[rdlf_env])
    # Determine which file to log to
    with restrict_umask(0o6):
        if log_file is not None:
            # We force creation to ensure proper permissions
            with log_file.open("a", encoding="utf8") as f:
                f.write("" if f.tell() == 0 else "\n")
        else:
            fd, fpath = mkstemp(suffix=".log")
            fd_close(fd)
            log_file = Path(fpath)
            atexit.register(_log_shutdown, log_file)
            if conf.debug:
                environ[rdlf_env] = str(log_file)
    # Setup logger
    fmt = CuteFormatter(**log.CF_KWARGS, colored=conf.colored)  # type: ignore[arg-type]
    fh = FileHandler(log_file, mode="a")
    stream = StreamHandler()
    root = getLogger()
    for i in (fh, stream):
        i.setFormatter(fmt)
        root.addHandler(i)
    # Set level
    lvl: int = log.level(conf.verbose)
    root.setLevel(lvl)
    root.info(
        "Logging level set to %s with colors %sABLED", getLevelName(lvl), "EN" if conf.colored else "DIS"
    )
    if reuse_log_file:
        root.warning("Reusing debug log file: %s", log_file)
    # Cleanup and return the log file
    return log_file


def serve(conf: ServerConfig, log_conf: LogConfig, favicon: Path) -> None:
    app.start(conf, _log_config(log_conf), favicon)



from dataclasses import dataclass, asdict, fields
from typing import TYPE_CHECKING
from urllib.parse import quote
from json import loads, dumps
from logging import getLogger
from hashlib import blake2s
from pathlib import Path

from human_readable import listing

from .errors import UsageError

if TYPE_CHECKING:
    from typing import Self


_CONFIG_LOG: str = "Config"
_CONFIG_FILE_PERMISSIONS: int = 0o600


@dataclass(init=False, slots=True)
class Result:
    """
    Result of a successful send/receive
    """

    checksum: blake2s | None
    total: int | None

    def __init__(self, total: bool, checksum: bool) -> None:
        self.checksum = blake2s(usedforsecurity=False) if checksum else None
        self.total = 0 if total else None


@dataclass(kw_only=True, frozen=True, slots=True)
class Config:
    """
    Information about where the remote pipe is
    """

    ssl: bool = True
    url: str = ""
    channel: str = ""
    password: str = ""
    timeout: int | None = None
    key_file: Path | None = None

    def __post_init__(self):
        if self.key_file and self.key_file.suffix == ".pub":
            getLogger(_CONFIG_LOG).warning("Signing key should be a private key: %s", self.key_file)

    def channel_url(self) -> str:
        return f"{self.url}/c/{quote(self.channel)}"

    @classmethod
    def keys(cls) -> list[str]:
        return [i.name for i in fields(cls)]

    @classmethod
    def load(cls, cli: dict[str, bool | str | Path | None], file: Path) -> Self:
        """
        Preference: CLI > Config file > Default
        """
        log = getLogger(_CONFIG_LOG)
        conf = asdict(Config())
        # Load config file then cli args
        if file.exists():
            file.chmod(_CONFIG_FILE_PERMISSIONS)  # Update permissions in case permissions are bad
            log.debug("Loading config file %s", file)
            conf.update(loads(file.read_text(encoding="utf-8")))
        else:
            log.warning("Config file does not exist: %s", file)
        conf.update({i: k for i, k in cli.items() if k is not None})
        # Finish
        if conf["key_file"] is not None:
            conf["key_file"] = Path(conf["key_file"])
        return cls(**conf)

    def save(self, file: Path) -> None:
        log = getLogger(_CONFIG_LOG)
        log.info("Mode: save-config")
        if not file.parent.exists():
            log.debug("Creating directory %s", file.parent)
            file.parent.mkdir(exist_ok=True)
        log.info("Saving config %s", self)
        if not file.exists():
            log.warning("Creating config file: %s", file)
            file.touch(mode=_CONFIG_FILE_PERMISSIONS)
            log.debug("Permissions set to: %s", oct(_CONFIG_FILE_PERMISSIONS))
        file.write_text(dumps(asdict(self), default=str), encoding="utf-8")
        file.chmod(_CONFIG_FILE_PERMISSIONS)  # Update permissions in case permissions are bad
        log.info("Config saved to %s", file)

    def validate(self) -> None:
        log = getLogger(_CONFIG_LOG)
        log.info("Validating config...")
        if missing := [i[1] for i in ((self.url, "URL"), (self.channel, "CHANNEL")) if not i[0]]:
            raise UsageError(f"Missing: {listing(missing, separator=",")}")
        if self.ssl and not self.url.startswith("https://"):
            raise UsageError(
                "SSL is required but URL does not start https scheme."
                " If raw http is desired, consider disabling SSL"
            )
        if self.key_file is not None and not self.key_file.exists():
            log.warning("Key file does not exist: %s", self.key_file)

    def __str__(self) -> str:
        d = asdict(self) | {"password": bool(self.password)}  # Do not leak password
        return "Config (with CLI overrides):\n  " + "\n  ".join(f"{i}: {k}" for i, k in d.items())

    def __repr__(self) -> str:
        return str(asdict(self))


# pylint: disable=too-many-instance-attributes
@dataclass(kw_only=True, frozen=True, slots=True)
class Mode:
    """
    Arguments used to decide how rpipe should operate
    Only one priority mode may be used at a time
    """

    # Priority modes
    print_config: bool
    update_config: bool
    outdated: bool
    server_version: bool
    blocked: bool
    query: bool
    # Recv/Send/Delete modes
    recv: bool
    send: bool
    delete: bool
    # Recv options
    block: bool
    peek: bool
    force: bool
    yes: bool
    # Send options
    file: Path | None  # None means sys.stdin if dir is None
    dir: Path | None
    ttl: int | None
    zstd: int | None
    threads: int
    # Recv / Send options
    encrypt: bool
    progress: bool | int
    total: bool
    checksum: bool

    @classmethod
    def keys(cls) -> list[str]:
        return [i.name for i in fields(cls)]

    def priority(self) -> bool:
        c = (
            self.print_config,
            self.update_config,
            self.outdated,
            self.server_version,
            self.blocked,
            self.query,
        ).count(True)
        assert c <= 1, "Sanity check on priority mode count failed"
        return c > 0



from logging import getLogger, Logger, DEBUG
from typing import TYPE_CHECKING
from dataclasses import fields

from flask import Response

if TYPE_CHECKING:
    from ...shared import UploadRequestParams, DownloadRequestParams
    from collections.abc import Callable
    from typing import TypeVar

    _ArgsT = TypeVar("_ArgsT", bound=Callable)


def log_params(log: Logger, p: UploadRequestParams | DownloadRequestParams) -> None:
    if not log.isEnabledFor(DEBUG):
        return
    log.debug("Request URL parameters:")
    for i in (k.name for k in fields(p)):
        log.debug("  %s: %s", i, getattr(p, i))


def log_response(log_name: str = "util"):
    """
    A decorator that logs the returned flask Responses to the log log_name
    """

    def decorator(func: Callable[[_ArgsT], Response]) -> Callable[[_ArgsT], Response]:
        def inner(*args, **kwargs) -> Response:
            ret: Response = func(*args, **kwargs)
            log = getLogger(log_name)
            if log.isEnabledFor(DEBUG):
                log.debug("Response:")
                log.debug("  Headers:")
                for i, k in ret.headers.items():
                    log.debug("    %s: %s", i, k)
                log.debug("  Status: %s", ret.status)
            return ret

        return inner

    return decorator



from typing import TYPE_CHECKING, cast
from collections import deque
import logging

from flask import request

from ...shared import WEB_VERSION, LFS, UploadResponseHeaders, UploadRequestParams, UploadEC
from ..util import MIN_VERSION, MAX_SIZE_HARD, MAX_SIZE_SOFT, plaintext
from .util import log_response, log_params
from ..server import Stream

if TYPE_CHECKING:
    from logging import Logger
    from flask import Response
    from ..server import State


DEFAULT_TTL: int = 300
_LOG = "write"


def _log_pipe_size(log: Logger, s: Stream) -> None:
    if log.isEnabledFor(logging.DEBUG):
        n = len(s)
        msg = "Pipe now contains %s/%s. It is %.2f%% full."
        log.debug(msg, LFS(n), LFS(s.capacity), 100 * n / s.capacity)


def _put_error_check(s: Stream | None, args: UploadRequestParams) -> Response | None:
    if s is None or s.id_ != args.stream_id:
        return plaintext("Stream ID mismatch.", UploadEC.conflict)
    if s.upload_complete:
        return plaintext("Cannot write to a completed stream.", UploadEC.forbidden)
    if args.version != s.version and not args.override:
        return plaintext(f"Override = False. Version should be: {s.version}", UploadEC.wrong_version)
    if s.full():
        return plaintext("Pipe full; wait for the downloader to download more.", UploadEC.wait)
    return None


# pylint: disable=too-many-return-statements
@log_response(_LOG)
def write(state: State, channel: str) -> Response:
    args = UploadRequestParams.from_dict(request.args)
    log = logging.getLogger(_LOG)
    log_params(log, args)
    # Version and size check
    if args.version != WEB_VERSION and (args.version < MIN_VERSION or args.version.invalid()):
        return plaintext(f"Bad version. Server requires >= {MIN_VERSION}", UploadEC.illegal_version)
    add = request.get_data()
    if len(add) > MAX_SIZE_HARD:
        return plaintext(f"Too much data sent. Max data size: {MAX_SIZE_SOFT}", UploadEC.too_big)
    # Starting a new stream, no stream ID should be present
    if request.method == "POST":
        if args.stream_id is not None:
            return plaintext("POST request should not have a stream_id", UploadEC.stream_id)
        new = Stream(
            data=deque([] if not add else [add]),
            ttl=DEFAULT_TTL if args.ttl is None else args.ttl,
            encrypted=args.encrypted,
            version=args.version,
            upload_complete=args.final,
        )
        headers = UploadResponseHeaders(stream_id=new.id_, max_size=MAX_SIZE_SOFT)
        with state as u:
            if (existing := u.streams.get(channel, None)) is not None and existing.locked:
                return plaintext("Channel is locked and cannot be edited.", UploadEC.locked)
            u.streams[channel] = new
            u.stats.write(channel)
        return plaintext("", 201, headers=headers.to_dict())
    # Continuing an existing stream, stream ID should be present
    if args.stream_id is None:
        return plaintext("PUT request missing stream id", UploadEC.stream_id)
    with state as unlocked:
        s: Stream | None = unlocked.streams.get(channel, None)
        if (err := _put_error_check(s, args)) is not None:
            return err
        if TYPE_CHECKING:
            s = cast(Stream, s)  # For type checker
        if s.locked:
            return plaintext("Channel is locked and cannot be edited.", UploadEC.locked)
        s.upload_complete = args.final
        if add:
            s.data.append(add)
            _log_pipe_size(log, s)
        if args.ttl is not None:
            s.ttl = args.ttl
        headers = UploadResponseHeaders(stream_id=s.id_, max_size=MAX_SIZE_SOFT)
    return plaintext("", 202, headers=headers.to_dict())



from typing import TYPE_CHECKING
from logging import getLogger
from json import dumps
import sys

from ...shared import TRACE, QueryEC, Version, version
from .errors import UsageError, VersionError, BlockedError
from .delete import delete
from .util import request
from .recv import recv
from .send import send

if TYPE_CHECKING:
    from pathlib import Path
    from .data import Result, Config, Mode


_LOG: str = "client"


def _print_config(conf: Config, config_file: Path) -> None:
    log = getLogger(_LOG)
    log.info("Mode: print-config")
    print(f"Path: {config_file}")
    print(conf)
    try:
        conf.validate()
        log.info("Config validated")
    except UsageError as e:
        log.warning("Config invalid %s", e)


def _check_outdated(conf: Config) -> None:
    log = getLogger(_LOG)
    log.info("Mode: Outdated")
    r = request("GET", f"{conf.url}/supported", timeout=conf.timeout)
    if not r.ok:
        raise RuntimeError(f"Failed to get server minimum version: {r}")
    info = r.json()
    log.info("Server supports clients: %s", info)
    ok = Version(info["min"]) <= version and all(version != Version(i) for i in info["banned"])
    print(f"{'' if ok else 'NOT '}SUPPORTED")


def _query(conf: Config) -> None:
    log = getLogger(_LOG)
    log.info("Mode: Query")
    if not conf.channel:
        raise UsageError("Channel unknown; try again with --channel")
    log.info("Querying channel %s ...", conf.channel)
    r = request("GET", f"{conf.url}/q/{conf.channel}", timeout=conf.timeout)
    log.debug("Got response %s", r)
    log.log(TRACE, "Data: %s", r.content)
    match r.status_code:
        case QueryEC.illegal_version:
            raise VersionError(r.text)
        case QueryEC.no_data:
            print(f"No data on channel: {conf.channel}")
            return
    if not r.ok:
        raise RuntimeError(f"Query failed. Error {r.status_code}: {r.text}")
    print(f"{conf.channel}: {dumps(r.json(), indent=4)}")


def _priority_actions(conf: Config, mode: Mode, config_file: Path) -> None:
    log = getLogger(_LOG)
    if mode.print_config:
        _print_config(conf, config_file)
        return
    if mode.update_config:
        log.info("Mode: Save Config")
        conf.save(config_file)
        return
    # Everything after this requires the URL
    if conf.url is None:
        raise UsageError("Missing: URL")
    # Remaining priority modes
    if mode.outdated:
        _check_outdated(conf)
    if mode.server_version or mode.blocked:
        log.info("Mode: Server Version")
        try:
            r = request("GET", f"{conf.url}/version", timeout=conf.timeout)
        except BlockedError:
            if mode.blocked:
                print("Server is blocking this IP address")
                sys.exit(1)
            raise
        if mode.blocked:
            print("Server is not blocking this IP address")
            return
        if not r.ok:
            raise RuntimeError(f"Failed to get version: {r}")
        print(f"rpipe_server {r.text}")
    if mode.query:
        _query(conf)


def rpipe(conf: Config, mode: Mode, config_file: Path) -> None:
    """
    rpipe: A remote piping tool
    :param conf: Configuration for the remote pipe, may be invalid at this point
    :param config_file: Path to the configuration file, may not exist
    :param mode: Mode to operate in, assumes flags are valid and within expected ranges
    """
    log = getLogger(_LOG)
    if mode.priority():
        _priority_actions(conf, mode, config_file)
        return
    conf.validate()
    if (mode.recv or mode.send) and not conf.password:
        log.warning("Encryption disabled: plaintext mode")
        if mode.zstd is not None:
            raise UsageError("Cannot compress data in plaintext mode")
    # Invoke mode
    log.info("HTTP timeout set to: %s", "DEFAULT" if conf.timeout is None else f"{conf.timeout} seconds")
    if mode.recv:
        rv: Result = recv(conf, mode)
    elif mode.send:
        rv = send(conf, mode)
    else:
        delete(conf)
        return
    # Print results
    if rv.checksum is not None:
        print(f"Blake2s: {rv.checksum.hexdigest()}", file=sys.stderr)
    if rv.total is not None:
        print(f"Total bytes {'sent' if mode.send else 'received'}: {rv.total}", file=sys.stderr)



from typing import TYPE_CHECKING
from json import dumps

from flask import Response

from ..shared import MAX_SOFT_SIZE_MIN, Version

if TYPE_CHECKING:
    from collections.abc import Sequence
    from enum import Enum


MIN_VERSION = Version("9.6.1")

# Maximum size of a request
# Note: The soft limit is soft to allow overhead of encryption headers and such
MAX_SIZE_SOFT: int = 64 * (1000**2)
assert MAX_SIZE_SOFT >= MAX_SOFT_SIZE_MIN
MAX_SIZE_HARD: int = 2 * MAX_SIZE_SOFT + 0x200  # For packets sent to the server only


def total_len(x: Sequence[bytes]) -> int:
    return sum(len(i) for i in x)


def plaintext(msg: str, status: Enum | int = 200, **kwargs) -> Response:
    """
    Return a plain text Response containing the arguments
    """
    code: int = status if isinstance(status, int) else status.value
    return Response(msg, status=code, mimetype="text/plain", **kwargs)


def json_response(js) -> Response:
    return Response(dumps(js, default=str), status=200, mimetype="application/json")



from threading import Thread, Condition
from logging import getLogger
import os

from ...shared import TRACE, LFS, SpooledTempFile, total_len


class IO:
    """
    A better version of stdin.read that doesn't hang as often
    Only meant to ever be used from a single thread at a time
    Will attempt to keep chunk bytes loaded at all times
    May use about an extra chunk bytes for stitching data together
    """

    __slots__ = ("_mlog", "_buffer", "_cond", "_eof", "_chunk", "_fd")

    def __init__(self, fd: int | SpooledTempFile, chunk: int) -> None:
        """
        :param fd: The file descriptor (or SpooledTemporaryFile) to read from
        :param chunk: The number of bytes to keep preloaded whenever possible
        """
        if isinstance(fd, SpooledTempFile) and not fd.is_virtual:
            fd = fd.fileno()  # This is a file on disk, so prefer reading as such
        self._fd: int | SpooledTempFile = fd
        self._mlog = getLogger("IO Main")
        self._buffer: list[bytes] = []
        self._cond = Condition()
        self._eof: bool = False  # Set when reader *thread* hits EOF; _buffer may still have data
        self._chunk: int = chunk
        if isinstance(self._fd, int):  # No need when reading from RAM
            self._mlog.info("Starting IO thread on fd %d with chunk size: %s", self._fd, LFS(self._chunk))
            Thread(target=self._worker, daemon=True).start()

    # Main Thread

    def increase_chunk(self, n: int) -> None:
        """
        Increase the chunk size to n bytes
        """
        if n < self._chunk:
            raise ValueError("Cannot decrease chunk size")
        if n != self._chunk:
            self._mlog.info("Updating IO chunk size to %s", LFS(n))
            with self._cond:
                self._chunk = n
                self._cond.notify()

    def read(self) -> tuple[bytes, bool]:
        """
        Read up to self._chunk bytes; b"" only if EOF
        :return: A tuple containing the bytes read and a boolean indicating EOF
        """
        with self._cond:
            if isinstance(self._fd, SpooledTempFile):
                return (ret := self._fd.read(self._chunk)), len(ret) == 0
            # If fd is a file descriptor
            self._cond.wait_for(lambda: self._buffer or self._eof)
            ret = b"".join(self._buffer)
            assert len(ret) <= self._chunk, "Sanity check failed"
            self._buffer.clear()
            self._cond.notify()
        if ret:
            self._mlog.debug("Read %s bytes of data", LFS(ret))
        return ret, self._eof and not self._buffer

    # Worker thread (only exists if fd is int)

    def _worker(self) -> None:
        """
        The worker thread that reads data from the file descriptor
        Always attempts to keep at least self._chunk bytes loaded
        """
        assert isinstance(self._fd, int), "Sanity check"
        log = getLogger("IO Thread")
        n = self._chunk
        until = lambda: total_len(self._buffer) < self._chunk
        # os.read may read in small chunks (ex. pipe buffer capacity in Linux)
        while data := os.read(self._fd, n):
            log.log(TRACE, "Loaded %s bytes of data from input", LFS(data))  # This can be spammy, so trace
            with self._cond:
                self._buffer.append(data)
                self._cond.notify()
                self._cond.wait_for(until)
                n = self._chunk - total_len(self._buffer)
        with self._cond:
            self._eof = True
            self._cond.notify()
        log.info("Read EOF. IO thread terminating.")
